// Lane.cpp: implementation of the Lane class.
//
//////////////////////////////////////////////////////////////////////

#include "Lane.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

	/**Default constructor*/
Lane::Lane()
{
	m_LeftLane = NULL;
	m_RightLane = NULL;
}

	/**Default destructor*/
Lane::~Lane()
{
	delete m_LeftLane;
	delete m_RightLane;
}

int Lane::m_index = 0;	// what's this for?

	/**Main constructor
	 *
	 * The main constructor used for creating a Lane object.
	 *
	 * @param RoadLength The length of the Road
	 * @param detectors The detectors to be used in the lane
	 * @param DirPos Whether the lane is in the positive direction or not
	 * @param AllowLaneChange Whether lane changing is allowed
	 */

Lane::Lane(int RoadLength, std::vector<Detector*> detectors, bool DirPos, bool AllowLaneChange)
{

	m_DirPos = DirPos;

	m_RoadLength = RoadLength;
	m_iLane = ++m_index;

	m_PrevVehiclePosition = SPACE_MIN;	// this is a property of IDM
	m_NoVeh = 0;

	m_vDetectors = detectors;

	m_AllowLaneChanging = AllowLaneChange;

	m_LeftLane = NULL;
	m_RightLane = NULL;
}

	/**Updates the positions of all Vehicles in the Lane
	 *
	 * This function is called for each iteration of the simulation, and handles the
	 * updating of the positions of Vehicles, and the removal of Vehicles which are
	 * no longer on the Road.
	 * If there are any Vehicles on the Road, the Vehicle at the head of the Lane is
	 * updated with respect to the clear Road ahead of it. Following this, all Vehicles 
	 * following the Vehicle at the head of the Road are updated with respect to the 
	 * preceding Vehicle. These Vehicles are also checked to see if it
	 * is time to determine if changing to another Lane would be advantageous.
	 * After all Vehicles have been updated, it is checked to see whether the Vehicle
	 * at the head of the Lane has yet passed the end of the Road. If it has, it is
	 * removed from the Lane and the simulation.
	 * Vehicles are also checked to see if they have passed a detector during the update
	 *
	 * @param step The timestep
	 * @return Whether the lane is empty or not
	 * @see changeLanes()
	 */

// returns laneEmpty = true if no vehicles in lane
bool Lane::update(double step)
{
	bool laneEmpty = true;
	if(m_pVehicles.size() > 0)
	{
		// Update the head of the lane to drive towards the end of the road
		double previousPos = m_pVehicles.front()->getPos();
		m_pVehicles.front()->update(step);
		double currentPos = m_pVehicles.front()->getPos();
		
		AddVehicleToDetector(previousPos, currentPos, m_pVehicles.front());

		// Update each vehicle based on the properties of the preceding vehicle
		for(int i = 1; i < m_pVehicles.size(); i++)
		{
			previousPos = m_pVehicles[i]->getPos();
			m_pVehicles[i]->update(step, m_pVehicles[i-1]);
			currentPos = m_pVehicles[i]->getPos();
			
			bool bChangedLane = false;
			if(m_AllowLaneChanging)											//Check if overtaking is allowed
			{
				bool bChangeStatus = m_pVehicles.at(i)->getChangeStatus();	// Ready to change lane?
				if(bChangeStatus)
				{
					bChangedLane = ChangeLanes(m_pVehicles.at(i));			// Has it changed?
					if(bChangedLane)
					{																			
						m_pVehicles.at(i)->setChangeStatus(false);			// Reset status
						m_pVehicles.erase(m_pVehicles.begin()+i);			// Remove from this lane
					}																			
				}
			}
			if(!bChangedLane)
				AddVehicleToDetector(previousPos, currentPos, m_pVehicles.at(i));
		} 

		m_PrevVehiclePosition = m_pVehicles.back()->getPos();

		//Remove any vehicles that are no longer on the road
		if(m_pVehicles.at(0)->getPos() > m_RoadLength)
		{
			delete m_pVehicles.front();
			m_pVehicles.erase(m_pVehicles.begin());
			m_NoVeh--;
		}
		laneEmpty = false;;
	}
	return laneEmpty;
}

	/** Adds a vehicle to the detectors list of vehicles for that step
	 *
	 * @param prevPosition The vehicle's previous position
	 * @param curPosition The vehicle's current position
	 * @param p
	 */
void Lane::AddVehicleToDetector(double prevPosition, double curPosition, Vehicle* pVeh)
{
	// if the vehicle has moved past it in the last time step
	if(prevPosition < m_OutputDetector->getPos() && curPosition >= m_OutputDetector->getPos())
		// add the vehicle to the detector
		m_OutputDetector->addVehicle(pVeh);
}

	/**Gets the length of the Lane
	 *
	 * @return The length of the Lane
	 */
	//Probably not needed?

double Lane::getLength()
{
	return m_RoadLength;
}

	/**Gets all the Vehicles currently in the Lane
	 *
	 * This function returns the vector which contains pointers to all
	 * Vehicles which are currently in the Lane
	 *
	 * @return All of the Vehicles in the Lane
	 */
std::vector<Vehicle*> Lane::getPos()
{
	return m_pVehicles;
}

	/**Sets a pointer to the Lane to the left of the current Lane
	 *
	 * @param left The Lane to point to
	 */
void Lane::setLeftLane(Lane *left)
{
	m_LeftLane = left;
}

	/**Sets a pointer to the Lane to the right of the current Lane
	 *
	 * @param right The Lane to point to
	 */
void Lane::setRightLane(Lane *right)
{
	m_RightLane = right;
}

int Lane::getNoVeh()
{
	return m_pVehicles.size();
}

double Lane::getLastPos()
{
	return m_PrevVehiclePosition;
}

int Lane::getIndex()
{
	return m_iLane;
}

////////////////////////////////////////////////////////////////
//////// BELOW HERE ARE THE LANE-CHANGING FUNCTIONS ////////////
////////////////////////////////////////////////////////////////

	/**Changes a Vehicle's Lane if advantageous and possible
	 *
	 * This function is called whenever it is time for a Vehicle to check whether
	 * or not it would be advantageous to move to another Lane.
	 *
	 */

bool Lane::ChangeLanes(Vehicle *pVeh)
{
	// since if it's a single lane you cannot change so let's not even bother
	if(m_RightLane == NULL && m_LeftLane == NULL)
		return false;

	bool DirPos = pVeh->getDirection();
	Vehicle* LeftFrontVehicle = NULL;	Vehicle* LeftBackVehicle = NULL;
	Vehicle* RightFrontVehicle = NULL;  Vehicle* RightBackVehicle = NULL;
	int iLeftLane = 0;					int iRightLane = 0;
	bool LeftLaneExists = m_LeftLane != NULL ? true : false;
	bool RightLaneExists = m_RightLane != NULL ? true : false;
	
	if(LeftLaneExists)
	{
		LeftFrontVehicle = FinadAdjacentVehicle(m_LeftLane, pVeh->getPos(), true, &iLeftLane);
		LeftBackVehicle = FinadAdjacentVehicle(m_LeftLane, pVeh->getPos(), false);
	}
	
	if(RightLaneExists)
	{
		RightFrontVehicle = FinadAdjacentVehicle(m_RightLane, pVeh->getPos(), true, &iRightLane);
		RightBackVehicle = FinadAdjacentVehicle(m_RightLane, pVeh->getPos(), false);
	}
	int LaneChangeID = 0;
	if(DirPos)	// in DirPos
		LaneChangeID = pVeh->DecideLaneChange(	LeftFrontVehicle,	LeftBackVehicle, 
												RightFrontVehicle,	RightBackVehicle,
												LeftLaneExists,		RightLaneExists);
	else	// in DirNeg so swap vehicles - see comment in Vehicle::DecideLaneChange()
		LaneChangeID = pVeh->DecideLaneChange(	RightFrontVehicle,	RightBackVehicle,
												LeftFrontVehicle,	LeftBackVehicle,
												RightLaneExists,	LeftLaneExists);

	if(LaneChangeID == 0)
		return false;
	else
	{
		ImplementLaneChange(pVeh, LaneChangeID, iLeftLane, iRightLane);
		return true;
	}
}

void Lane::ImplementLaneChange(Vehicle* pVeh, int LaneChangeID, int iVehLeftLane, int iVehRightLane)
{
	// we're in the business of changing lanes so let's prepare
	Lane* VehiclesLeftLane;		Lane* VehiclesRightLane;
	bool DirPos = pVeh->getDirection();

	if(DirPos)
	{
		VehiclesLeftLane = m_LeftLane;
		VehiclesRightLane = m_RightLane;
	}
	else
	{
		VehiclesLeftLane = m_RightLane;
		VehiclesRightLane = m_LeftLane;
		//Swapping vehicle index in new Lane
		int temp = iVehLeftLane;
		iVehLeftLane = iVehRightLane;
		iVehRightLane = temp;
	}
	if(LaneChangeID == 1)
	{
		// change to the vehicle's left lane
		pVeh->setLane(VehiclesLeftLane->getIndex());
		VehiclesLeftLane->insert(iVehLeftLane, pVeh);
	}
	else
	{
		// change to vehicle's right lane
		pVeh->setLane(VehiclesRightLane->getIndex());
		VehiclesRightLane->insert(iVehRightLane, pVeh);
	};
}

// this function finds the vehicle in the adjacent lane,
// in front of (front = true) the location or behind (front = false)
// note iAdjacentLane is optional and returns the position
Vehicle* Lane::FinadAdjacentVehicle(Lane* AdjacentLane, double location, bool front, int* iAdjacentLane)
{
	std::vector<Vehicle*> AdjacentLaneVeh = AdjacentLane->getPos();

	if(AdjacentLaneVeh.size() == 0)
		return NULL;

	// searches from back of nextLane up to the vehicle in front
	int i = AdjacentLaneVeh.size() - 1;

	while(i >= 0 && AdjacentLaneVeh.at(i)->getPos() < location)
		i--;

	if(iAdjacentLane != NULL)	// i points to the vehicle in front of us in the other lane
		*iAdjacentLane = i+1;	// place the position in iAdjacentLane

	//We need to check for bounds here, since we're returning a Vehicle, not an index

	if(front && i >= 0)
		return AdjacentLaneVeh.at(i);
	else if(front && i < 0)
		return NULL;
	else if(!front && i == AdjacentLaneVeh.size() - 1)
		return NULL;
	else
		return AdjacentLaneVeh.at(i+1);
}

	/**Inserts a Vehicle into the Lane
	 *
	 * This function inserts a given Vehicle into a particular position in
	 * a Lane
	 *
	 * @param i The position to insert the Vehicle into
	 * @param pVeh The Vehicle to insert
	 */
void Lane::insert(int i, Vehicle *pVeh)
{
	if(i == m_pVehicles.size())
	{
		if(m_PrevVehiclePosition >= SPACE_MIN)			// this checked before this ufnciton is called anyway
		{
			m_pVehicles.push_back(pVeh);
			m_PrevVehiclePosition = pVeh->getPos();
		}
	}

	else if(i <= 0)
		m_pVehicles.insert(m_pVehicles.begin(), pVeh);
	
	else
		m_pVehicles.insert(m_pVehicles.begin()+i, pVeh);

	m_NoVeh++;
}

void Lane::setOutputDetector(Detector *out)
{
	m_OutputDetector = out;
}
