
#include "Road.h"

	/** Constructor */
Road::Road()
{
	m_bEndOfFile = false;
}

	/** Destructor */
Road::~Road()
{

}

	/** Updates the properties of the road
	 *
	 * @param step The timestep
	 * @return Whether or not the simulation is still running
	 *
	 * This function is called for each timestep of the simulation and updates its
	 * directions, once the directions are not empty
	 */
bool Road::update(double step,const double curTime)
{
	bool SimulationOver = false;
	bool RoadEmpty = false;
	bool DirPosEmpty = false;
	bool DirNegEmpty = false;

	// since there's always one direction
	DirPosEmpty = DirectionPos.update(step, curTime);

	if(m_NoDirections == 2)
		DirNegEmpty = DirectionNeg.update(step, curTime);

	if(DirPosEmpty && DirNegEmpty)
		RoadEmpty = true;

	if(RoadEmpty && m_bEndOfFile)	// if no more vehicles and road is empty
		SimulationOver = true;		// we are finished!

	return SimulationOver;			
}

	/** Initialises the road
	 *
	 * @return The time to the first vehicle
	 *
	 * This function initialises the road object. It begins by loading a buffer
	 * of vehicles which it will process, and maps these vehicles to their 
	 * respective lanes. It then takes account of cumulative lane identifiers by
	 * loading each direction's lanes into a vector of cumulative lanes.
	 * Detectors are also set in each direction to facilitate the retrieval of
	 * updated vehicle information when a vehicle passes a specified point. 
	 */
double Road::init()
{
	int i;
	m_BufferVehicles = m_FileHandler->loadTruckGroup(INPUT_FILE_BUFFER_SIZE);
	for(i = 0; i < m_BufferVehicles.size(); i++)
		MapTrafLaneToSimLane(m_BufferVehicles.at(i));

	// Since we'll always have at least one direction
	// true means positive x-direction
	Detector* pDetDirPos = new Detector(m_PosOutputDetectorDirPos, true, m_FileHandler);
	DirectionPos.init(true, pDetDirPos, m_AllowLaneChanging);			
	DirectionPos.createLanes(m_RoadLength, m_NoLanesDirPos, m_vDetectorsPos);
	
	// Using cumulative lane identifiers
	for(i = 0; i < DirectionPos.getNoLanes(); i++){
		m_vLanes.push_back(&DirectionPos.getLane(i));
	}
	
	if(m_NoDirections == 2)
	{
		// false means the negative x-direction
		Detector* pDetDirNeg = new Detector(m_PosOutputDetectorDirNeg, false, m_FileHandler);
		DirectionNeg.init(false, pDetDirNeg, m_AllowLaneChanging);		
		DirectionNeg.createLanes(m_RoadLength, m_NoLanesDirNeg, m_vDetectorsNeg);

		for(i = 0; i < DirectionNeg.getNoLanes(); i++){
			m_vLanes.push_back(&DirectionNeg.getLane(i));
		}
	}

	// Return the time to that of the first truck of the traffic file
	return m_BufferVehicles.front()->getTime();
}

	/** Populates the road
	 *
	 * @param step The timestep
	 * @param CurrentTime The simulation's current time
	 *
	 * This function handles the population of the road which is called in each
	 * timestep. The main loop handles the processing of the vehicles in the buffer.
	 * If the head of the buffer is currently entering the road, then the other 
	 * vehicles are also checked. However, since the vehicles are sorted in order of
	 * time, if the head of the buffer is not currently entering the road, then no
	 * other vehicle is either.
	 * When a vehicle enters the road from the buffer, the next vehicle is read from
	 * the file being used, and added to the buffer.
	 */
void Road::populate(double step, const double curTime)
{
	bool frontOn = true;

	while(m_BufferVehicles.size() > 0 && frontOn)
	{
		frontOn = false;
		int lane = m_BufferVehicles.front()->getLane() - 1;

		// Using cumulative lane identifiers
		int noVeh = m_vLanes.at(lane)->getNoVeh();	// crashes here if vehicle is in lane = 3 say if there are only 2 lanes
													// i.e. if user has put in fewer lanes than are in the truck file.
													// we need to allow for this.

		// set the driver
		setIDMDriverModel(m_BufferVehicles.front());

		if(m_vLanes.at(lane)->getLastPos() >= SPACE_MIN)			//If there is space for another vehicle
		{
			if(m_BufferVehicles.front()->getTime() <= curTime)	//If it is time for the next vehicle
			{
				m_vLanes.at(lane)->insert(noVeh, m_BufferVehicles.front());
				m_BufferVehicles.erase(m_BufferVehicles.begin());

				Truck* tempVeh = m_FileHandler->readLine();
				if(tempVeh != NULL)
				{
					MapTrafLaneToSimLane(tempVeh);
					m_BufferVehicles.push_back(tempVeh);
				}
				else
					m_bEndOfFile = true;
				frontOn = true;
			}
		}

	}
}

// This function updates the vehicle's original traffic file
// lane assignment, to an appropriate simulation lane

	/** Maps vehicles to appropriate lanes
	 *
	 * @param pVeh The vehicle to map
	 *
	 * This function updates the lane provided by the vehicle's original traffic
	 * file assignment to an appropriate lane in the simulation. This is to
	 * facilitate the simulation of more lanes than the input file allows for
	 */
void Road::MapTrafLaneToSimLane(Vehicle* pVeh)
{
	if(!pVeh->getDirection())	// only if it's in the negative direction
	{
		int TrafFileNoLanes = m_TrafFileNoLanesDirPos + m_TrafFileNoLanesDirNeg;
		int lanesFromNegVerge = TrafFileNoLanes - pVeh->getLane();
		int newLane = m_NoLanes - lanesFromNegVerge;
		pVeh->setLane(newLane);
	}
}

	/** Sets a vehicle's drivermodel
	 *
	 * @param pVeh The vehicle to set
	 *
	 * This function sets the drivermodel of a particular vehicle based on the
	 * type of that vehicle
	 */
void Road::setIDMDriverModel(Vehicle* pVeh)
{
	WORD VEH_ID = pVeh->getID();
	
	switch(VEH_ID)
	{
		case VEH_ID_CAR:
			pVeh->setDriver( m_pIDMParams_Car->Generate() );
			break;
		case VEH_ID_SMALLTRUCK:
			pVeh->setDriver( m_pIDMParams_SmallTruck->Generate() );
			break;
		case VEH_ID_LARGETRUCK:
			pVeh->setDriver( m_pIDMParams_LargeTruck->Generate() );
			break;
		case VEH_ID_CRANE:
			pVeh->setDriver( m_pIDMParams_Crane->Generate() );
			break;
		case VEH_ID_LOWLOADER:
			pVeh->setDriver( m_pIDMParams_Lowloader->Generate() );
			break;
		default:
			pVeh->setDriver( m_pIDMParams_Car->Generate() );
	};
}

	/** Gets all the vehicles on the road
	 *
	 * @return All the vehicles on the road
	 *
	 */
M2D Road::getVehicles()
{
	// This function needs to be able to cater for two directions
	// and for an unequal number of lanes
	
	M2D pos;

	M2D temp = DirectionPos.getGlobalPos();	// get DirPos positions
	for(int i = 0; i < m_NoLanesDirPos; i++)		// and add lane by lane
		pos.push_back(temp[i]);

	if(m_NoDirections == 2)		// Uni-directional roads are in Dir1
	{
		temp.clear();
		
		temp = DirectionNeg.getGlobalPos();	// get DirNeg positions
		for(i = 0; i < m_NoLanesDirNeg; i++)		// and add lane by lane
			pos.push_back(temp[i]);
	}

	return pos;
}

	/** Gets the length of the road
	 *
	 * @return The length of the road
	 */
double Road::getLength()
{
	return m_RoadLength;
}

	/** Sets the number of lanes in the positive direction
	 *
	 * @param nlpos The number of lanes in the positive direction
	 */
void Road::setNoLanesDirPos(int nlpos)
{
	m_NoLanesDirPos = nlpos;
}

	/** Sets the number of lanes in the negative direction
	 *
	 * @param nlneg The number of lanes in the negative direction
	 */
void Road::setNoLanesDirNeg(int nlneg)
{
	m_NoLanesDirNeg = nlneg;
}

	/** Sets the number of directions
	 *
	 * @param nd The number of directions
	 */
void Road::setNoDirections(int nd)
{
	m_NoDirections = nd;
}

	/** Sets the number of lanes
	 *
	 * @param nl The number of lanes
	 */
void Road::setNoLanes(int nl)
{
	m_NoLanes = nl;
}

	/** Sets the length of the road
	 *
	 * @param L The length of the road
	 */
void Road::setRoadLength(int L)
{
	m_RoadLength = L;
}

	/** Initialises the file handler
	 *
	 * @param inFile The input file to read from
	 * @param outFile The output file to write to
	 * @param fileType The type of file
	 *
	 * This function determines whether the type of file being used is
	 * CASTOR or SAFT format, and creates an appropriate filehandler
	 * based on the input provided.
	 */
void Road::initTruckGroup(string inFile, string outFile, WORD fileType)
{
	if(fileType == CASTOR)
	{
		m_FileHandler = new CASTORFile(inFile, outFile);

	}
	else if(fileType == SAFT)
	{
		m_FileHandler = new SAFTFile(inFile, outFile);
	}
}

	/** Sets the Car IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_Car(CIDMParameterSet* Params)
{
	m_pIDMParams_Car = Params;
}

	/** Sets the Small Truck IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_SmallTruck(CIDMParameterSet* Params)
{
	m_pIDMParams_SmallTruck = Params;
}

	/** Sets the Large Truck IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_LargeTruck(CIDMParameterSet* Params)
{
	m_pIDMParams_LargeTruck = Params;
}

	/** Sets the Crane IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_Crane(CIDMParameterSet* Params)
{
	m_pIDMParams_Crane = Params;
}

	/** Sets the Lowloader IDM Parameter Set
	 *
	 * @param Params The IDM Parameter Set
	 */
void Road::setIDMParams_Lowloader(CIDMParameterSet* Params)
{
	m_pIDMParams_Lowloader = Params;
}

	/** Sets the detector properties
	 *
	 * @param detectors The vector of detectors to use
	 */
void Road::setDetectors(std::vector<Detector*> detectors)
{
/*	for(int i = 0; i < detectors.size(); i++)
	{
		Detector* d = new Detector(pos.at(i), 0, 1);
		m_vDetectors.push_back(d);
	}
	*/
}

	/** Sets the number of lanes in the positive direction in the traffic file
	 *
	 * @param nl The number of lanes
	 */
void Road::setTrafFileNoLanesDirPos(int nl)
{
	m_TrafFileNoLanesDirPos = nl;
}

	/** Sets the number of lanes in the negative direction in the traffic file
	 *
	 * @param nl The number of lanes
	 */
void Road::setTrafFileNoLanesDirNeg(int nl)
{
	m_TrafFileNoLanesDirNeg = nl;
}

	/** Sets the position of output detector in the positive direction
	 *
	 * @param loc The position of output detector
	 */
void Road::setPosOuputDetectorDirPos(int loc)
{
	m_PosOutputDetectorDirPos = loc;
}

	/** Sets the position of output detector in the negative direction
	 *
	 * @param loc The position of output detector
	 *
	 */
void Road::setPosOuputDetectorDirNeg(int loc)
{
	// change from overall road coordinates to lane coordinates
	m_PosOutputDetectorDirNeg = m_RoadLength - loc;
}

	/** Gets whether or not lane changing is allowed
	 *
	 * @return Whether or not lane changing is allowed
	 */
bool Road::getAllowLaneChanging()
{
	return m_AllowLaneChanging;
}

	/** Sets whether or not lane changing is allowed
	 *
	 * @param status Whether or not lane changing is allowed
	 */
void Road::setAllowLaneChanging(bool status)
{
	m_AllowLaneChanging = status;
}
