#include "Vehicle.h"


	/** Default Constructor */
Vehicle::Vehicle(){

}

	/** Default Destructor */
Vehicle::~Vehicle(){

}

	/** Updates the position of the Vehicle based on the timestep
	 * @param step the timestep
	 *
	 * The Vehicle calling this function will always be the head of traffic in
	 * its particular Lane, as it has no Vehicle in front to reference and instead
	 * drives towards the end of the Road
	 * This function also handles the timer for when the Vehicle will check to see if changing Lane
	 * would be advantageous.
	 */

void Vehicle::update(double step)
{
	if( changeTime(step) )
		setChangeStatus(true);

	double accel = calcAccel();

	updateProperties(step, accel);
}

	/** Updates the position of the Vehicle based on the timestep and preceding Vehicle
	 * @param step the timestep
	 * @param pV a pointer to the Vehicle in front of this one
	 *
	 * Every Vehicle in the Lane will call this function except the head of the Lane. The velocity
	 * and acceleration of the Vehicle will be altered based on the velocity of, and distance to,
	 * the Vehicle in front.
	 * This function also handles the timer for when the Vehicle will check to see if changing Lane
	 * would be advantageous.
	 */

void Vehicle::update(double step, Vehicle *pV){
	bool change = changeTime(step);

	if(change)
		setChangeStatus(true);

	double accel = calcAccel(pV);

	updateProperties(step, accel);
}

	/** Calculates a Vehicle's acceleration when unimpeded by other Vehicles
	 *
	 * This function serves the purpose of determining the acceleration of a Vehicle when
	 * there are no other Vehicles in front of it in the same Lane. Based on the distance
	 * to the end of the Road, the DriverModel alters acceleration accordingly
	 */
double Vehicle::calcAccel()
{
	m_pDriver->setVel(vel);

	//MAGIC NUMBER: ROAD END
	double accel = m_pDriver->update(vel, 12500-m_x);

	return accel;
}

	/** Calculates a Vehicle's acceleration based on properties of the preceding Vehicle
	 * @param pV A pointer to the preceding Vehicle
	 *
	 * This function serves the purpose of determining the acceleration of a Vehicle when
	 * there is another Vehicle in front of it in the same Lane. Based on the distance
	 * between the two vehicles, and the difference in their velocity, the DriverModel
	 * alters acceleration accordingly.
	 */
double Vehicle::calcAccel(Vehicle *pV)
{
	m_pDriver->setVel(vel);

	double s = pV->getPos() - pV->getLength() - m_x;

	double delta_v = vel - pV->getVelocity();
	double accel = m_pDriver->update(delta_v, s);

	return accel;
}

	/** Updates the velocity, acceleration and position of the Vehicle based its acceleration and the timestep
	 * @param step The timestep
	 * @param accel The Vehicle's acceleration
	 *
	 * This function updates the properties of the Vehicle based on how much it has accelerated in the previous
	 * timestep, assuming average velocity over the timestep. The velocity, and acceleration of the Vehicle are
	 * updated, as well as the Vehicle's position.
	 */

void Vehicle::updateProperties(double step, double accel)
{
	// Assume average velocity over timestep
	double v_old = vel;
	double v_new = vel + accel*step;
	
	m_x += 0.5*(v_old + v_new)*step;
	roadPos = m_x;
	vel = v_new;

	acc = accel;

	// zero speeds may occur when very close to stationary obstacle
	// eliminate as per Treiber's comments to Carpool Tunnel
	if(vel < 0)
		vel = 0;
}

	/** Sets the velocity of the Vehicle to a given value
	 * @param velocity the Vehicle's new velocity
	 */

void Vehicle::setVelocity(double velocity){
	vel = velocity;
}

	/** Sets the length of the Vehicle to a given value
	 * @param length The Vehicle's new length
	 */

void Vehicle::setLength(double length){
	len = length;
}

	/** Sets the acceleration of the Vehicle to a given value
	 * @param acceler the Vehicle's new acceleration
	 */

void Vehicle::setAccel(double acceler){
	acc = acceler;
}

	/** Sets the lane that the Vehicle is in
	 * @param l the Vehicle's lane
	 */

void Vehicle::setLane(int l){
	lane = l;
}

	/** Sets the direction of the Vehicle
	 * @param DirPointsRight Whether the vehicle is in a positive or negative direction
	 */

void Vehicle::setDirection(bool DirPos)
{
	m_DirPos = DirPos;
}


	/** Sets the position of the Vehicle to a given value
	 * @param p the Vehicle's position
	 */

void Vehicle::setPos(double p){
	m_x = p;
}

	/** Sets whether or not the Vehicle wishes to change lanes
	 * @param stat whether or not the Vehicle wants to change lanes
	 */

void Vehicle::setChangeStatus(bool stat){
	laneChange = stat;
}

	/** Sets the position of the Vehicle on the road. For graphical use
	 * @param r the Vehicle's position on the road
	 */

void Vehicle::setRoadPos(double r){
	roadPos = r;
}


	/** Gets the Vehicle's position
	 * @return m_x the Vehicle's position
	 */

double Vehicle::getPos(){
	return m_x;
}

	/** Gets the desired velocity of the Vehicle's driver model
	 * @return the driver model's desired velocity
	 */

double Vehicle::getDesiredVel(){
	return m_pDriver->getDesiredVel();
}

	/** Gets the position of the Vehicle on the road. Graphical use
	 * @return roadPos the position of the Vehicle on the road
	 */

double Vehicle::getRoadPos(){
	return roadPos;
}

	/** Gets the class of the Vehicle
	 * @return ID the Vehicle's class
	 */

int Vehicle::getID(){
	return ID;
}

	/** Gets whether or not the Vehicle wishes to change lane
	 * @return laneChange whether or not the Vehicle wishes to change lane
	 */

bool Vehicle::getChangeStatus(){
	return laneChange;
}

	/** Gets the velocity of the Vehicle
	 * @return vel the velocity of the Vehicle
	 */

double Vehicle::getVelocity(){
	return vel;
}

	/** Gets the length of the Vehicle
	 * @return len the length of the Vehicle
	 */

double Vehicle::getLength(){
	return len;
}

	/** Gets the acceleration of the Vehicle
	 * @return acc the acceleration of the Vehicle
	 */

double Vehicle::getAccel(){
	return acc;
}

	/** Gets the lane that the Vehicle is in
	 * @return lane the Vehicle's lane
	 */

double Vehicle::getLane(){
	return lane;
}

	/** Gets the direction that the Vehicle is travelling in
	 * @return dir the Vehicle's direction
	 */

bool Vehicle::getDirection()
{
	return m_DirPos;
}

	/** Handles the timer for checking Lane changes
	 *
	 * @param step The timestep
	 * @return Whether or not it is time to check for a Lane change
	 */
bool Vehicle::changeTime(double step)
{
	tdelay += step;

	if (tdelay>=T_DELAY){

		tdelay=tdelay-T_DELAY;
		return true;

	}
	else {return false;}
}

	/**Gets the Vehicle's DriverModel
	 * @return the Vehicle's driver
	 */
DriverModel* Vehicle::getDriver()
{
	 return m_pDriver;
}

	/**Sets the Vehicle's DriverModel
	 * @param driver The Vehicle's driver
	 */
void Vehicle::setDriver(IDM driver)
{
	m_IDMDriver = driver;
	m_pDriver = &m_IDMDriver;
}

	/** Default implementation for printing CASTOR data
	 * @param pTruck The string to print the information to
	 */
void Vehicle::doCASTORData(char* pTruck, const double curTime)
{

}
	/** Default implementation for printing SAFT data
	 * @param pTruck The string to print the information to
	 */
void Vehicle::doSAFTData(char *pTruck, const double curTime)
{

}

////////////////////////////////////////////////////////
////////////////// LANE CHANGING FUNCTIONS /////////////
////////////////////////////////////////////////////////

	/** Handles whether or not the Vehicle will change Lane
	 *
	 * @param LeftFrontVehicle The Vehicle in front in the Left Lane
	 * @param LeftBackVehicle The Vehicle behind in the Left Lane
	 * @param RightFrontVehicle The Vehicle in front in the Right Lane
	 * @param RightBackVehicle The Vehicle behind in the Right Lane
	 * @param LeftLaneExists Whether or not there is a Lane to the left
	 * @param RightLaneExists Whether or not there is a Lane to the Right
	 * @return The Lane to change to
	 * @see LaneChangeAdvantage()
	 *
	 * This function determines which Lane the Vehicle should change to if it changes
	 * Lane at all. Pointers to Vehicles that would be in front of, and behind, the 
	 * current Vehicle if it were in either the right or left Lane are passed to the
	 * function, along with identifiers which specify whether or not there exists a 
	 * Lane to the left or right of the current Lane.
	 */
int Vehicle::DecideLaneChange(	Vehicle* LeftFrontVehicle,	Vehicle* LeftBackVehicle,
								Vehicle* RightFrontVehicle, Vehicle* RightBackVehicle,
								bool LeftLaneExists,		bool RightLaneExists)
{
	// Note that a possible conflict between what is the lane's leftLane and rightLane
	// and the Vehicle's leftLane and Rightlane. For DirPos they are the same, but not 
	// for DirNeg. e.g.
	// DirPos - Lane 1: leftLane = 2: rightLane = 0;
	// DirNeg - Lane 5: leftLane = 6: rightLane = 4
	// Whereas for the Vehicle:
	// DirPos - in lane 1: lane to left = 2: lane to right = 0;
	// DirNeg - in lane 5: lane to left = 4: lane to right = 6;
	// Therefore:
	// THIS FUNCTION NAMES LANES AS VIEWED BY THE VEHICLE
	
	double LeftLaneAdvantage = 0.0;		double RightLaneAdvantage = 0.0;

	if(DRIVE_ON_RIGHT)
	{	// we are overtaking properly
		if(LeftLaneExists)
			LeftLaneAdvantage = LaneChangeAdvantage(LeftFrontVehicle, LeftBackVehicle, true);
		// or we are being bold
		if(RightLaneExists)
			RightLaneAdvantage = LaneChangeAdvantage(RightFrontVehicle, RightBackVehicle, false);
	}
	else
	{	// we're boldly undertaking
		if(LeftLaneExists)
			LeftLaneAdvantage = LaneChangeAdvantage(LeftFrontVehicle, LeftBackVehicle, false);
		// we are overtaking properly
		if(RightLaneExists)
			RightLaneAdvantage = LaneChangeAdvantage(RightFrontVehicle, RightBackVehicle, true);
	}

	const double zero = 0.0001;
	if(LeftLaneAdvantage < zero && RightLaneAdvantage < zero)
		return 0;									// do not change lane
	else if(LeftLaneAdvantage > RightLaneAdvantage)
		return 1;									// move to the left lane
	else
		return 2;									// move to the right lane
}

	/**Gets the advantage a Vehicle would gain from switching Lanes
	 *
	 * @param FrontVehicle The Vehicle in front in the new Lane
	 * @param BackVehicle The Vehicle behind in the new Lane
	 * @param overtake Whether the Vehicle is overtaking or undertaking
	 * @return The net advantage of the Lane change
	 *
	 * This function takes pointers to the Vehicles both in front, and behind, the position 
	 * that the current Vehicle would have in the new Lane. If the Vehicle that would be in 
	 * front is null, then the change in acceleration is decided with relation to acceleration
	 * on an open Road. Otherwise, the acceleration is decided with relation to the Vehicle in
	 * front.
	 */
double Vehicle::LaneChangeAdvantage(Vehicle* FrontVehicle, Vehicle* BackVehicle, bool overtake)
{
	// These are IDM properties and may be reomved from Constants.h
	double GapToFront = MIN_GAP;			double GapToBack = MIN_GAP; 
	double CurrentBackAccel	= SAFE_BRAKING;	double ProposedBackAccel= SAFE_BRAKING;
	double FrontChangeAccel	= 0.0;

	// calculate the above lane change parameters
	if(FrontVehicle != NULL)
	{
		GapToFront = FrontVehicle->getPos() - FrontVehicle->getLength() - this->getPos();
		FrontChangeAccel = this->calcAccel(FrontVehicle) - this->getAccel();
	}
	else
		FrontChangeAccel = this->calcAccel() - this->getAccel();	// no front Vehicle

	if(BackVehicle != NULL)
	{
		GapToBack = this->getPos() - this->getLength() - BackVehicle->getPos();
		ProposedBackAccel = BackVehicle->calcAccel(this);
		// Since the acceleration of the back Vehicle at the next step
		// is affected by the presence of the Vehicle in front of it
		// we must take that into account and not an open-road acceleration
		// but only if there is a Vehicle in front!
		if(FrontVehicle != NULL)
			CurrentBackAccel = BackVehicle->calcAccel(FrontVehicle);
		else
			CurrentBackAccel = BackVehicle->getAccel();
	}

	double advantage =  m_pDriver->LaneChange(GapToFront, GapToBack, FrontChangeAccel, 
												CurrentBackAccel, ProposedBackAccel, overtake);
	return advantage;
}

